/*  sjm2mml.c:  main & convert functions for sjm2mml
    Takeshi H.
    copyleft  2002-2004
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sjm2mml.h"

/*  global variables  */
char filename[FNLEN+1];
FILE *fpout;                /*  output file pointer  */
struct st_whole_sjm sjm;
char str_note[STRNOTELEN+1];
struct st_flags flag;
struct st_mem_flags mem_flags = {
    bufstat:    No,
    sjmbuf:     No,
    notes:      No,      
    velocity:   No,  
    instrument: No,
    title:      No,     
};

int main(int argc, char **argv)
{
    int onkai;
    int nagasa;
    int tr;    /*  track counter  */
    char track_symbol;  /*  track symbol  */
    int ch_no;  /*  channel number written in MML file  */
//    int orig_inst, orig_drum_inst;
    extern char inst_drums[], inst_origmelos[], inst_origdrums[];
    int note;  /*  each note on the track  */
    char *drum_def_file, *originst_def_file, *origdrum_def_file;
    FILE *fp;
    char *p;
    
    void parse_options(int argc, char *argv[]);
    extern FILE* search_file(char *filename);
    extern enum InstrumentCode parse_instrument( char inst );

    /*  parse options  */
    parse_options(argc, argv);

    if (flag.help_exit == Yes) {
        fprintf(stderr, "Ringing Melody Converter SJM to MML: Ver. %s \n", SJM2MML_VERSION);
        fprintf(stderr, "usage: %s [options] infile.sjm\n", argv[0]);
        fputs( "   options:\n"                                     , stderr );
        fputs( "   -h, -v       :  print help, exit\n"             , stderr );
        fputs( "   -ke          :  keep empty tracks\n"            , stderr );
        fputs( "   -bd,-nd      :  use built-in drum defaults, skip definition file. \n"    , stderr );
        fputs( "   -bom         :  use built-in original melody defaults, skip definition file. \n"    , stderr );
        fputs( "   -bod         :  use built-in original drum defaults, skip definition file. \n"    , stderr );
        fputs( "   -o <outfile> :  output file name is <outfile>\n", stderr );
        /*  fputs( "   -oo          :  same as -o, <outfile> is infile.mml\n", stderr );  */
        fputs( "      <outfile> is overwritten if exists\n"        , stderr );

        return Help;        /*  exit with exit_code Help  */
    }

    /*  load data into memory  */
    sjm.sh = (sjm_header *) load_sjm_file(filename, &sjm);
        /*  We can access header information with sh later  */


    /*  load drum No. data into memory  */
    if ( flag.read_drumdef == Yes ){
        //  file name is able to be specified by user in the futre
        drum_def_file=DEFAULT_DRUMINSTFILE;
#ifdef DEBUG
        fprintf(stderr, "drum difinition file: %s\n", drum_def_file);
#endif /*  DEBUG  */
        if ( ( fp=search_file( drum_def_file ) ) != NULL ){
            read_def_file(fp, inst_drums, INST_DRUMSET_START, INST_DRUMS);
            fclose(fp);
        } else {
#ifdef DEBUG
            fprintf(stderr, "Couldn't find %s anywhere, falling back to built-in default.\n", drum_def_file);

#endif  /*  DEBUG  */
        }
        /*
        set_drum_number();
        */
    } else {
#ifdef DEBUG
        fputs("I'm to use built-in drum difinition.\n", stderr);
#endif /*  DEBUG  */
    }

    /*  load original melody No. data into memory  */
    if ( flag.read_origmelodef == Yes ){
        //  file name is able to be specified by user in the futre
        originst_def_file=DEFAULT_ORIGINSTFILE;
#ifdef DEBUG
        fprintf(stderr, "original instrument difinition file: %s\n", originst_def_file);
#endif /*  DEBUG  */
        if ( ( fp=search_file( originst_def_file ) ) != NULL ){
            read_def_file(fp, inst_origmelos, INST_ORIGINAL_INST_START, INST_ORIGINAL_INSTS);
            fclose(fp);
        } else {
#ifdef DEBUG
            fprintf(stderr, "Couldn't find %s anywhere, falling back to built-in default.\n", originst_def_file);

#endif  /*  DEBUG  */
        }
    } else {
#ifdef DEBUG
        fputs("I'm to use built-in original melody difinition.\n", stderr);
#endif /*  DEBUG  */
    }

    /*  load original drum No. data into memory  */
    if ( flag.read_origdrumdef == Yes ){
        //  file name is able to be specified by user in the futre
        origdrum_def_file=DEFAULT_ORIGDRUMINSTFILE;
#ifdef DEBUG
        fprintf(stderr, "original drum difinition file: %s\n", origdrum_def_file);
#endif /*  DEBUG  */
        if ( ( fp=search_file( origdrum_def_file ) ) != NULL ){
            read_def_file(fp, inst_origdrums, INST_ORIGINAL_DRUM_START, INST_ORIGINAL_DRUMS);
            fclose(fp);
        } else {
#ifdef DEBUG
            fprintf(stderr, "Couldn't find %s anywhere, falling back to built-in default.\n", origdrum_def_file);

#endif  /*  DEBUG  */
        }
    } else {
#ifdef DEBUG
        fputs("I'm to use built-in original drum difinition.\n", stderr);
#endif /*  DEBUG  */
    }


    /*  get header info & complement structure */
    get_header_info(&sjm);

    /*  put MML file header  */
        /*  Title of song  */
    fprintf( fpout, "#title %c%s%c\n" ,'"', sjm.title ,'"');
        /*  timebase is fixed because 48 is enough for SJM  */
    fputs( "#timebase 48\n", fpout );
        /*  record the converter version  */
    fprintf( fpout, ";converted by sjm2mml ver. %s\n",SJM2MML_VERSION );

    /*  get notes in each track  */
    p=sjm.melody_top;

    track_symbol='A';
    ch_no = 1;
    for( tr=0; tr< ((sjm_header *)sjm.sjmbuf)->num_of_tracks; tr++){

#ifdef DEBUG
        fprintf (stderr, "track %d\n",tr);
#endif

        /*  if track length is 0, skip it  */
        if ( flag.keep_emptytrack == No  && sjm.notes[tr] == 0 )
            continue;

        /*  Here should be a check of number of track ... beyond 'Z'?  */
        if ( track_symbol > 'Z' ){
            fprintf (stderr, "track_symbol exceed 'Z'.\n");
            fprintf (stderr, "sorry, I can't handle so many tracks.\n");
            clean_exit( TooManyTracks );
        }

        /*  Track Name  :  main track starts with 'A', ....  */
        fprintf( fpout, "\n%c ", track_symbol );
        track_symbol++;

        /*  parse_instrument  */
        switch ( parse_instrument( sjm.instrument[tr] ) ) {

            case NormalInst:  /*  normal instrument  */

                /*  Channel number  */
                if ( ch_no == DRUM_CHANNEL )
                    ch_no++;  /*  avoid Drum channel  */
                fprintf( fpout, "C%d ", ch_no );
                ch_no++;

                /*  instrumntal number  */
                fprintf( fpout, "@%d ", sjm.instrument[tr] );

                break;

            case NormalDrum:  /*  Drum set  */

                /*  Channel number  */
                fprintf( fpout, "C%d ", DRUM_CHANNEL );  /*  drumset is on ch 10  */

                fprintf( fpout, "J%d ", 
                  inst_drums[ sjm.instrument[tr] - INST_DRUMSET_START ] );

                break;
                
            case OriginalInst:  /*  Original Tones  */

                /*  Channel number  */
                if ( ch_no == DRUM_CHANNEL )
                    ch_no++;  /*  avoid Drum channel  */
                fprintf( fpout, "C%d ", ch_no );
                ch_no++;

                /*  Instrument number  */
                fprintf( fpout, "@%d ", 
                    inst_origmelos[ 
                        sjm.instrument[tr] - INST_ORIGINAL_INST_START
                    ]
                );

                break;

            case OriginalDrum:  /*  Original Drums  */

                /*  Channel number  */
                /*  usually, drumset is on ch 10  */
                fprintf( fpout, "C%d ", DRUM_CHANNEL ); 

                /*  Drum Inst Number  */
                fprintf( fpout, "J%d ", 
                    inst_origdrums[
                      sjm.instrument[tr] - INST_ORIGINAL_DRUM_START 
                    ]
                );

                break;

            default:
                break;
        }


        /*  Tempo is written only on the 1st track  */
        if ( tr == 0) fprintf( fpout, "t%d ", sjm.tempo );

        /*  Velocity, in fact intensity is written on the track  */
        fprintf( fpout, "k%d ", sjm.velocity[tr]);

        /*  each note on the track  */
        for( note =0; note< sjm.notes[tr]/2; note++){
            onkai=*p++;
            fputs( pitch_of_note( str_note,  onkai), fpout );

            nagasa=*p++;
            fputs( length_of_note(str_note, nagasa), fpout );
        }

    }

    fputs( "\n", fpout );

    /*  close output file pointer  */
    if (flag.stdout == No)
        fclose(fpout);

    /*  Free() buffers  */
    free(sjm.title);
    mem_flags.title = No;
    free(sjm.instrument);
    mem_flags.instrument = No;
    free(sjm.velocity);
    mem_flags.velocity = No;
    free(sjm.notes);
    mem_flags.notes = No;
    free(sjm.sjmbuf);
    mem_flags.sjmbuf = No;


    return Normal;
}

void parse_options(int argc, char *argv[])
{
    int i;
    enum YesNo have_sjm_file = No;

    /*  flag defaults  */
    flag.help_exit = No;
    flag.read_drumdef = Yes;
    flag.read_origmelodef = Yes;
    flag.read_origdrumdef = Yes;
    flag.keep_emptytrack = No;
    flag.stdout = Yes;

    if(argc == 1){  /*  no argument  */
        flag.help_exit = Yes;
        return ;
    } else {  /*  at least 1 argument  */
        for ( i =1; i < argc; i++){
            if ( strcmp( argv[i], "-h") == SAME ){
                flag.help_exit = Yes;
                return ;
            } else if ( strcmp( argv[i], "-v") == SAME ){
                /*  show version and exit  */
                flag.help_exit = Yes;
            /*  } else if ( strcmp( argv[i], "-oo") == SAME ){  */
                /*  output file name is infile.mml  */



            } else if ( strcmp( argv[i], "-o") == SAME ){
                /*  output file name  */
                ++i;
                fpout = fopen( argv[i],"w" );
                if ( fpout != NULL ){
                    flag.stdout = No;
                } else {
                    fprintf( stderr, "Can't open file %s for output.\n", argv[i] );
                    clean_exit( FileOpenFailure );
                }

            } else if ( strcmp( argv[i], "-bd" ) == SAME 
                     || strcmp( argv[i], "-nd" ) == SAME /*  old  */){
                /*  use built-in drums definition ?  */
                /*  doesn't read drums.def file? (old) */
                flag.read_drumdef = No;
            } else if ( strcmp( argv[i], "-bom" ) == SAME ){
                /*  use built-in original melody definition ?  */
                flag.read_origmelodef = No;

            } else if ( strcmp( argv[i], "-bod" ) == SAME ){
                /*  use built-in original drums definition ?  */
                flag.read_origdrumdef = No;


            } else if ( strcmp( argv[i], "-ke") == SAME ){
                /*  keep empty track(s)?  */
                flag.keep_emptytrack = Yes;
            } else {
                /*  decide .sjm filename  */
                if ( have_sjm_file == No ) {
                    strncpy(filename, argv[i], FNLEN);
                    have_sjm_file = Yes;
                } else {
                    fprintf( stderr, "Too many .sjm file. I already have %s\n", filename );
                    clean_exit( TooManySJM );
                }
            }

        }

/*  
        if ( ){
        }
        fpout = fopen( ,"w" );
        if ( fpout != NULL ){
            flag.stdout = No;
        } else {
            fprintf( stderr, "Can't open file %s for output.\n", argv[i] );
            clean_exit( FileOpenFailure );
        }
*/
        if ( flag.stdout == Yes ){
            fpout = stdout;
        }
    }
}


/*  When error occured, free() memories allocated by malloc()  */
void clean_exit( enum ecode error){
    if (mem_flags.bufstat == Yes)
        free(sjm.bufstat);

    if (mem_flags.sjmbuf == Yes)
        free(sjm.sjmbuf);

    if (mem_flags.notes == Yes)
        free(sjm.notes);

    if (mem_flags.velocity == Yes)
        free(sjm.velocity);

    if (mem_flags.instrument == Yes)
        free(sjm.instrument);

    if (mem_flags.title == Yes)
        free(sjm.title);

    if (flag.stdout == No)
        fclose(fpout);

    exit( error );
}
