/*  sjmfile.c:  file loading functions for sjm2mml.
        .sjm file:  melody file
        .def file:  instrument assign file
    Takeshi H.
    copyleft  2002-2004
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "sjm2mml.h"

sjm_header *load_sjm_file(char *sjmfilename, whole_sjm *sjmp)
{
    FILE *fp;
    char *p;
    int bufsize;   /*  buffer size = .sjm file size  */
    int i;         /*  loop counter  */


#ifdef DEBUG
    fputs(sjmfilename, stderr);
    fputs("\n", stderr);
#endif

    fp = fopen(sjmfilename,"r");
    if (fp == NULL){
        fputs(" No such .sjm file!\n" ,stderr);
        clean_exit( FileOpenFailure );
    }

    /*  Allocate memory for struct stat  */
    sjmp -> bufstat = (struct stat *)malloc( sizeof(struct stat));
    if (sjmp -> bufstat != NULL){
        /*  Mark the malloc flag  */
        mem_flags.bufstat = Yes;
    } else {
        fputs("Malloc Error at sjmp -> bufstat.\n" ,stderr);
        clean_exit( MallocError );
    }

    /*  Get file size  */
    stat(sjmfilename, sjmp -> bufstat);
    bufsize = sjmp -> bufstat -> st_size;
#ifdef DEBUG
    fprintf(stderr, "bufsize =%d\n",bufsize);
#endif

    free(sjmp -> bufstat);
    mem_flags.bufstat = No;

    /*  Allocate buffer for sjm data  */
    sjmp -> sjmbuf = (char *)malloc(bufsize);
    if (sjmp -> sjmbuf != NULL){
        /*  Mark the malloc flag  */
        mem_flags.sjmbuf = Yes;
    } else {
        fputs("Malloc Error for sjmbuf.\n" ,stderr);
        clean_exit( MallocError );
    }

    /*  Read from sjm file, store in buffer  */
    for (p=sjmp -> sjmbuf, i=0; i < bufsize; p++, i++) {
        *p = fgetc(fp);
    }

    return (sjm_header *)(sjmp->sjmbuf);
}


/*  search filename and try to open
    1. current dir  
    2. USERCONFDIR in home dir
*/
FILE* search_file(char *filename)
{
    char def_filename[FNLEN];
    char *homedirname;
    FILE *fp = NULL;

    if ( strlen(filename) <= FNLEN ) {
        strcpy ( def_filename, filename );
//        strncpy ( def_filename, filename, strlen(filename) );
        if ( ( fp=fopen(def_filename, "r") ) != NULL ){
#ifdef DEBUG
            fprintf(stderr, "Succeed to open %s \n", def_filename );
#endif  /*  DEBUG  */
        } else {
#ifdef DEBUG
            fprintf(stderr, "Failed to open %s \n", def_filename );
#endif  /*  DEBUG  */
            /*  not found.  */
            /*  next, try in in $HOME/.sjm2mml directory  */
            if ( ( homedirname = getenv("HOME") ) != NULL) {
                if ( strlen(homedirname) + 1 
                     + strlen( USERCONFIGDIR ) +1
                     + strlen(filename) 
                    <= FNLEN ) {

                    strcpy( def_filename, homedirname );
                    strcat( def_filename, "/" );
                    strcat( def_filename, USERCONFIGDIR );
                    strcat( def_filename, "/" );
                    strcat( def_filename, filename );

                    if ( ( fp=fopen(def_filename, "r") ) != NULL ){
#ifdef DEBUG
                        fprintf(stderr, "Succeed to load %s \n", def_filename );
#endif  /*  DEBUG  */
                    }
                } else {
                    fputs( ".def file path is too long!\n", stderr);
                    clean_exit( TooLongString );
                }
            } else {
#ifdef DEBUG
                fputs("No HOME environment variable.\n", stderr);
#endif  /*  DEBUG  */
            }
        }
    } else {
        fputs( ".def file path is too long!\n", stderr);
        clean_exit( TooLongString );
    }

    return fp;
}

/*  read .def file and store value to array  */
void read_def_file(FILE *fp, char *array, int start_element, int arraysize)
{
    int line=0;
    char lineptr[DEFLINELEN];
    int midiinst, sjminst;

    /*  process file   */
    while( fgets(lineptr, DEFLINELEN, fp) != (char *)NULL){
        line++;
        if ( *lineptr != INST_DRUM_COMMENT ){
            sscanf(lineptr, "%d = %d ; %*s", &sjminst, &midiinst );
#ifdef DEBUG
    fputs(lineptr,stderr);
#endif
            
            if ( start_element <= sjminst 
                && sjminst < start_element + arraysize ){
                array[ sjminst - start_element ] = midiinst ;
            } else {
            /*  Error: sjminst is invalid value  */
                fprintf( stderr, "Invalid Instrument: line %d\n", line);
            }
        }
    }
#ifdef DEBUG
    fprintf(stderr, "read_def_file: processed %d lines.\n",line);
#endif
}

